from typing import Union
from base_class import Calculator
from exceptions.negative_root import NegativeRootError

# Definition of custom type for clearer type hinting
Number = Union[int, float]

class HandlerCalculator(Calculator):
    """
    Simple calculator which perfoms various mathematical operations.

    The main difference with it's parent -Calculator- class is that instead of
        stopping execution when an exception is raised, the memory value is not
        changed and continues operation. Otherwise, it works exactly the same.
    
    For more in-depth explanation, refer to help() or .__doc__ for the
        specific method. Only the default behaviour is described here.

        
    Attributes
    ----------
        memory : Number
            the current memory state of the calculator instance (default 0)

    Methods
    -------
        add(*args, only=False)
            adds the arguments to the current memory value

        subtract(*args, only=False, reverse=False)
            substracts the arguments from the current memory value
            
        multiply(*args, only=False, factor=False)
            multiplies the values passed and the current memory

        divide(*args, only=False, reverse=False)
            divides the current memory by the values passed as arguments

        nth_root(n=1, value=0)
            calculates the (n)th root of the current memory

        power(p=1, value=0)
            elevates the current memory to the power of p

        reset(value=0)
            resets the calculator memory to 0

        get_memory()
            returns the current value in memory

        screen(message='')
            prints on stdout a pretty version of the calculator's memory

    Notes
    -----
        'Number' is an alias for Union[int, float]
        All methods parse the arguments to float type and work in sequential
            order, starting with the current memory value by default.
    """

    def __init__(self, memory: Number=0) -> None:
        """
        HandlerCalculator instance constructor (inherited from Calculator)

        Parameters
        ----------
            memory : Number, optional
                Initializing value for the memory of the calculator

        Returns
        -------
            None

        Notes
        -----
            If an invalid type is passed, HandlerCalculator is initialized with
                memory = 0
        """

        try:
            super().__init__(memory)
        except TypeError as te:
            print(f"add() failed with TypeError:\
                \nmessage: {str(te)}\
                \ninitialized with memory = 0\n")
            self.memory = 0


    def add(self, *args: Number, only: bool=False) -> float:
        """
        Parses and adds the values of args to calculator.memory by default

        Parameters
        ----------
            args : Number
                The values to add
            only : bool, optional
                If True, add() performs 'only' the addition of the args

        Returns
        -------
            float
                The result of the operation if all can be completed, else, the
                initial memory state

        Notes
        -----
            If no argument is passed, no changes in memory occur
            If the full operation is not completed successfully (i.e. one or
                some of the arguments is not valid) no changes in memory occur

        Examples
        --------
            (all starting with instance.memory = 10)

            >>> instance.add(1, 2, 3) == 10+1+2+3
            13.0
            >>> instance.add(1, 2, 3, only=True) == 1+2+3
            6.0
            >>> instance.add(1, 2, "x") <- Catches ValueError and resets
            10.0
        """

        try:
            super().add(*args, only=only)
        except ValueError as ve:
            print(f"add() failed with ValueError:\
                \nmessage: {str(ve)}\
                \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"add() failed with TypeError:\
                \nmessage: {str(te)}\
                \nno operations were performed, memory is {self.memory}\n")
        
        return self.memory
    
    
    def subtract(self, *args: Number,
                 only: bool=False, reverse: bool=False) -> float:
        """
        Parses and subtracts the values of args from calculator.memory from
            left to right starting by the memory value by default

        Parameters
        ----------
            args : Number
                The values to subtract
            only : bool, optional
                If True, it performs 'only' the subtraction of args
            reverse : bool, optional
                If True, subtracts the values from right to left

        Returns
        -------
            float
                The result of the operation or the previous memory state if the
                operation fails to complete

        Notes
        -----
            If no argument is passed, no changes in memory occur
            If the full operation is not completed successfully (i.e. one or
                some of the arguments is not valid) no changes in memory occur

        Examples
        --------
            (All starting with instance.memory = 5)

            >>> instance.subtract(1, 2, 3) == 5-3-2-1
            -1.0
            >>> instance.subtract(1, 2, 3, only=True) ==  1-2-3
            -4.0
            >>> instance.subtract(1, 2, 3, reverse=True) == 3-2-1-5
            -5.0
            >>> instance.subtract(1, 2, 3, reverse=True, only=True) == 3-2-1
            0.0
            >>> instance.subtract(1, 2, "should be ValueError")
            5.0
        """

        try:
            super().subtract(*args, only=only, reverse=reverse)
        except ValueError as ve:
            print(f"subtract() failed with ValueError:\
                  \nmessage: {str(ve)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"subtract() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nno operations were performed, memory is {self.memory}\n")
            
        return self.memory
    
    
    def multiply(self, *args: Number, 
                 only=False, factor=False) -> float:
        """
        Parses and multiplies the values of args in sequential order, starting
            by the current memory by default

        Parameters
        ----------
            args : Number
                The values to multiply
            only : bool, optional
                If True, returns the product of all numbers without considering
                    the value stored in memory. Multiplies 'only' the arguments
            factor : bool, optional
                If set to True, multiplies each args value by the memory and
                    then multiplies the results together

        Returns
        -------
            float
                The product of the given numbers or the previous memory value
                if the operation fails to complete
        
        Notes
        -----
            'factor' will override 'only'
            If no arguments are passed, no change in memory occur
            If the full operation is not completed successfully (i.e. one or
                some of the arguments is not valid) no changes in memory occur

        Examples
        --------
            All assuming instance.memory = 5

            >>> instance.multiply(1, 2, 3) == 5*3*2*1
            30.0
            >>> instance.multiply(1, 2, 3, only=True) == 1*2*3
            6.0
            >>> instance.multiply(1, 2, 3, factor=True) == (5*1)*(5*2)*(5*3)
            750.0
            >>> instance.multiply(1, 2, [10, 20, 30])
            5.0
        """

        try:
            super().multiply(*args, only=only, factor=factor)
        except ValueError as ve:
            print(f"multiply() failed with ValueError:\
                  \nmessage: {str(ve)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"multiply() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nno operations were performed, memory is {self.memory}\n")
            
        return self.memory
    
    
    def divide(self, *args: Number,
               only: bool=False, reverse: bool=False) -> float:
        """
        Divides the values in sequential order, memory value taken as first
            dividend
        
        Parameters
        ----------
            args : Number
                The values to use as divisors
            only : bool, optional
                If True, divides the values in args only: args[0] is first
                    dividend, args[1] is first divisor
            reverse : bool, optional
                If True, divides the values from right to left

        Returns
        ------
            float
                the result of the operation or the previous memory state if the
                operation fails to complete in full
                
        Notes
        -----
            If no argument is passed, no changes in memory occur
            If the full operation is not completed successfully (i.e. one or
                some of the arguments is not valid) no changes in memory occur

        Examples
        --------
            (All starting with instance.memory = 5)

            >>> instance.divide()
            5
            >>> instance.divide(10)
            0.5
            >>> instance.divide(10, reverse=True)
            2.0
            >>> instance.divide(10, 2, only=True)
            5.0
            >>> instance.divide(10, 2, only=True, reverse=True)
            0.2
            >>> instance.divide(9, 0, [])
            5

        """
        try:
            super().divide(*args, only=only, reverse=reverse)
        except ValueError as ve:
            print(f"divide() failed with ValueError:\
                  \nmessage: {str(ve)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"divide() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except ZeroDivisionError as zde:
            print(f"divide() failed with ZeroDivisionError:\
                  \nmessage: {str(zde)}\
                  \nno operations were performed, memory is {self.memory}\n")
            
        return self.memory
    
    
    def nth_root(self, n: int=1, value: Number= 0) -> float:
        """
        Parses n to int and calculates the n-th root of the value in memory
            if no other parameter is specified

        Parameters
        ----------
            n : int
                The degree of the root
            value : Number, optional
                The radicant of the root; instance.memory by default

        Returns
        ------
            float
                the result of the operation or the previous memory if an error
                    occured
                
        Notes
        -----
            If no arguments are passed, no changes in memory occur

        Examples
        --------
            (All starting with instance.memory = 5)

            >>> instance.nth_root()
            5.0
            >>> instance.nth_root(2)
            2.23606797749979
            >>> calculator.nth_root(2, 16)
            4.0

            >>> instance.reset(-3)
            >>> instance.nth_root(2)
            5.0
        """

        try:
            super().nth_root(n, value)
        except ValueError as ve:
            print(f"nth_root() failed with ValueError:\
                  \nmessage: {str(ve)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"nth_root() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except ZeroDivisionError as zde:
            print(f"nth_root() failed with ZeroDivisionError:\
                  \nmessage: {str(zde)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except NegativeRootError as nnr:
            print(f"nth_root() failed with NegativeRootError:\
                  \nmessage: {str(nnr)}\
                  \nno operations were performed, memory is {self.memory}\n")
            
        return self.memory
    

    def power(self, p: int=1, value: Number=0) -> float:
        """
        Removes decimal part from p and elevates the value in memory to the
            power of p

        Parameters
        ----------
            p : int
                The exponent of the power
            value : Number, optional
                The base of the power; instance.memory by default

        Raises
        ------
            ValueError
                If the argument is non-valid (e.g. "spam")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)

        Returns
        ------
            float
                The result of the operation or 1 if 0⁰
                If an exception is caught, the original memory value is returned
                
        Notes
        -----
            If no arguments are passed, no changes in memory occur

        Examples
        --------
            (All starting with instance.memory = 5)

            >>> instance.power(2)
            25
            >>> instance.power(2, 10)
            100
            >>> instance.power("a")
            5

            >>> instance.reset(-5)
            >>> instance.power(3)
            -125
        """

        try:
            super().power(p, value)
        except ValueError as ve:
            print(f"power() failed with ValueError:\
                  \nmessage: {str(ve)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except TypeError as te:
            print(f"power() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nno operations were performed, memory is {self.memory}\n")
        except ZeroDivisionError as zde:
            print(f"power() failed with ZeroDivisionError:\
                  \nmessage: {str(zde)}\
                  \nno operations were performed, memory is {self.memory}\n")
            
        return self.memory
    
    
    def reset(self, value: Number= 0) -> None:
        """
        Resets the calculator memory to 0 or the specified value

        Parameters
        ----------
            value : Number, optional
                The value to set the memory to. Default is 0

        Notes
        -----
            If the argument passed is not valid, memory is set to 0

        Returns
        ------
            None
        """

        try:
            super().reset(value)
        except TypeError as te:
            self.memory = 0
            print(f"reset() failed with TypeError:\
                  \nmessage: {str(te)}\
                  \nmemory was set to 0\n")
        


from typing import Union
from exceptions.negative_root import NegativeRootError

# Definition of custom type for clearer type hinting
Number = Union[int, float]

class Calculator:
    """
    Simple calculator which perfoms various mathematical operations.

    For more in-depth explanation, refer to help() or .__doc__ for the
        specific method. Only the default behaviour is described here.
    
        
    Attributes
    ----------
        memory : Number
            the current memory state of the calculator instance (default 0)

    Methods
    -------
        add(*args, only=False)
            adds the arguments to the current memory value

        subtract(*args, only=False, reverse=False)
            substracts the arguments from the current memory value

        multiply(*args, only=False, factor=False)
            multiplies the values passed and the current memory

        divide(*args, only=False, reverse=False)
            divides the current memory by the values passed as arguments

        nth_root(n=1, value=0)
            calculates the (n)th root of the current memory

        power(p=1, value=0)
            elevates the current memory to the p power

        reset(value=0)
            resets the calculator memory to 0

        get_memory()
            returns the current value in memory

        screen(message='')
            prints on stdout a pretty version of the calculator's memory

    Notes
    -----
        'Number' is an alias for Union[int, float]
        All methods parse the arguments to float type and work in sequential
            order, starting with the current memory value by default.
    """

    def __init__(self, memory: Number=0) -> None:
        """
        Calculator instance constructor

        Parameters
        ----------
            memory : Number, optional
                Initializing value for the memory of the calculator

        Raises
        ------
            TypeError
                If the memory is set to other than int or float types

        Returns
        -------
            None

        """

        if not isinstance(memory, (int, float)):
            raise TypeError("memory must be set to a number")
        self.memory = memory


    def add(self, *args: Number, only: bool=False) -> float:
        """
        Parses and adds the values of args to calculator.memory by default

        Parameters
        ----------
            args : Number
                The values to add
            only : bool, optional
                If True, add() performs 'only' the addition of the args

        Raises
        ------
            ValueError
                If the argument is non-valid (e.g. "hello")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)

        Returns
        -------
            float
                The result of the operation

        Notes
        -----
            If no argument is passed, no changes in memory occur

        Examples
        --------
            (all starting with calculator.memory = 10)

            >>> calculator.add(1, 2, 3) == 10+1+2+3
            13.0
            >>> calculator.add(1, 2, 3, only=True) == 1+2+3
            6.0
            >>> calculator.add(1, 2, "x")
            ValueError
        """

        if only:
            initial: float = 0
        else:
            initial = self.memory

        for value in args:
            initial += float(value)

        self.memory = initial
        return self.memory
    
    
    def subtract(self, *args, only: bool=False, reverse: bool=False) -> float:
        """
        Parses and subtracts the values of args from calculator.memory from
            left to right starting by the memory value by default

        Parameters
        ----------
            args : Number
                The values to subtract
            only : bool, optional
                If True, it performs 'only' the subtraction of args
            reverse : bool, optional
                If True, subtracts the values from right to left

        Raises
        ------
            ValueError
                If the argument is a non-valid string (e.g. "hello")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)

        Returns
        -------
            float
                The result of the operation

        Notes
        -----
            If no argument is passed, no changes in memory occur

        Examples
        --------
            (All starting with calculator.memory = 5)

            >>> calculator.subtract(1, 2, 3) == 5-3-2-1
            -1.0
            >>> calculator.subtract(1, 2, 3, only=True) ==  1-2-3
            -4.0
            >>> calculator.subtract(1, 2, 3, reverse=True) == 3-2-1-5
            -5.0
            >>> calculator.subtract(1, 2, 3, reverse=True, only=True) == 3-2-1
            0.0
            >>> calculator.subtract(1, 2, [3, 4, 5])
            TypeError
        """

        arg_list = list(args)
        # Setting up the initial value
        if reverse:
            arg_list.reverse()      # Reversing the list
            
        if reverse or only:
            initial = arg_list[0]   # Using the first element as initial
            arg_list.pop(0)         # Removing it from the list
        else:
            initial = self.memory   # Else, first element is the memory

        if reverse:
            for value in arg_list:
                initial -= float(value)
            if not only:
                initial -= self.memory
        else:
            for value in arg_list:
                initial -= float(value)

        self.memory = initial
        return self.memory
    
    
    def multiply(self, *args: Number, only=False, factor=False) -> float:
        """
        Parses and multiplies the values of args in sequential order, starting
            by the current memory by default

        Parameters
        ----------
            args : Number
                The values to multiply
            only : bool, optional
                If True, returns the product of all numbers without considering
                    the value stored in memory. Multiplies 'only' the arguments
            factor : bool, optional
                If set to True, multiplies each args value by the memory and
                    then multiplies the results together

        Raises
        ------
            ValueError
                If the argument is a non-valid string (e.g. "R2D2")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)

        Returns
        -------
            float
                The product of the given numbers 
        
        Notes
        -----
            'factor' will override 'only'
            If no arguments are passed, no change in memory occur

        Examples
        --------
            All assuming calculator.memory = 5

            >>> calculator.multiply(1, 2, 3) == 5*3*2*1
            30.0
            >>> calculator.multiply(1, 2, 3, only=True) == 1*2*3
            6.0
            >>> calculator.multiply(1, 2, 3, factor=True) == (5*1)*(5*2)*(5*3)
            750.0
            >>> calculator.multiply(1, 2, [10, 20, 30])
            TypeError
        """

        arg_list = list(args)
        if factor:
            arg_list = [self.memory * float(value) for value in arg_list]
        if only or factor:
            initial = arg_list[0]
            arg_list.pop(0)
        else:
            initial = self.memory

        for value in arg_list:
            initial *= float(value)
        
        self.memory = initial
        return self.memory
    
    
    def divide(self, *args: Number,
               only: bool=False, reverse: bool=False) -> float:
        """
        Divides the values in sequential order, memory value taken as first
            dividend
        
        Parameters
        ----------
            args : Number
                The values to use as divisors
            only : bool, optional
                If True, divides the values in args only: args[0] is first
                    dividend, args[1] is first divisor
            reverse : bool, optional
                If True, divides the values from right to left

        Raises
        ------
            ValueError
                If the argument is non-valid (e.g. "spam")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)
            ZeroDivisionError
                Pretty self explanatory, isn't it? 😅

        Returns
        ------
            float
                the result of the operation
                
        Notes
        -----
            If no argument is passed, no changes in memory occur

        Examples
        --------
            (All starting with calculator.memory = 5)

            >>> calculator.divide()
            5
            >>> calculator.divide(10)
            0.5
            >>> calculator.divide(10, reverse=True)
            2.0
            >>> calculator.divide(10, 2, only=True)
            5.0
            >>> calculator.divide(10, 2, only=True, reverse=True)
            0.2
            >>> calculator.divide(10, 2, [])
            TypeError
        """

        arg_list = list(args)
        if reverse:
            arg_list.reverse()
            
        if reverse or only:
            initial = float(arg_list[0])
            arg_list.pop(0)
        else:
            initial = float(self.memory)

        if reverse:
            for value in arg_list:
                initial /= float(value)
            if not only:
                initial /= self.memory
        else:
            for value in arg_list:
                initial /= float(value)

        self.memory = initial
        return self.memory
    

    def nth_root(self, n: int=1, value: Number=0) -> float:
        """
        Parses n to int and calculates the n-th root of the value in memory
            if no other parameter is specified

        Parameters
        ----------
            n : int
                The degree of the root
            value : Number, optional
                The radicant of the root; calculator.memory by default

        Raises
        ------
            ValueError
                If the argument is non-valid (e.g. "spam")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)
            ZeroDivisionError
                If trying to find the 0th root
            NegativeRootError
                If trying to calculate the root of a negative number. This
                    calculator hasn't been designed to handle complex numbers

        Returns
        ------
            float
                the result of the operation
                
        Notes
        -----
            If no arguments are passed, no changes in memory occur
            If memory is a negative number and no arguments ar passed, it will
                throw a NegativeRootException

        Examples
        --------
            (All starting with calculator.memory = 5)

            >>> calculator.nth_root()
            5.0
            >>> calculator.nth_root(2)
            2.23606797749979
            >>> calculator.nth_root(2, 16)
            4.0

            >>> calculator.reset(-3)
            >>> calculator.nth_root(2)
            NegativeRootError
        """

        value = float(value) or self.memory
        n = int(n)
        if value < 0:
            raise NegativeRootError(n, value)
        
        self.memory = value ** (1/n)
        return self.memory
    

    def power(self, p: int=1, value: Number=0) -> float:
        """
        Removes decimal part from p and elevates the value in memory to the
            power of p

        Parameters
        ----------
            p : int
                The exponent of the power
            value : Number, optional
                The base of the power. calculator.memory by default

        Raises
        ------
            ValueError
                If the argument is non-valid (e.g. "spam")
            TypeError
                If the argument is a non-supported type (e.g. list or tuple)
            ZeroDivisionError
                If trying to raise 0 to a negative value

        Returns
        ------
            float
                the result of the operation or 1 if 0⁰
                
        Notes
        -----
            If no arguments are passed, no changes in memory occur

        Examples
        --------
            (All starting with calculator.memory = 5)

            >>> calculator.power(2)
            25
            >>> calculator.power(2, 10)
            100
            >>> calculator.power("a")
            ValueError

            >>> calculator.reset(-5)
            >>> calculator.power(3)
            -125
        """

        power: float = 1
        base: float = value or self.memory
        p = int(p)

        if p == 0 and base == 0:
            self.memory = 1
            return self.memory
        elif p > 0:
            for _ in range(p):
                power *= base
        else:
            p = p * (-1)
            for _ in range(p):
                power *= base
            power = 1 / power

        self.memory = power
        return self.memory
    

    def reset(self, value: Number=0) -> None:
        """
        Resets the calculator memory to 0 or the specified value

        Parameters
        ----------
            value : Number, optional
                The value to set the memory to. Default is 0

        Raises
        ------
            TypeError
                If the argument is a non-supported type (e.g. string or list)

        Returns
        ------
            None
        """

        if not isinstance(value, (int, float)):
            raise TypeError("memory must be reset to a number")
        else:
            self.memory = value

    def get_memory(self) -> float:
        """
        Method to save the current memory value as a variable

        Parameters
        ----------
            None

        Returns
        ------
            calculator.memory : float
        """

        return self.memory
    

    def screen(self, message: str='') -> None:
        """
        Prints a prettified version of the memory value or a message of your
            liking in a screen-like format taking 32 characters wide by 3 lines

        Parameters
        ----------
            message : str, optional
                The message to display (truncated to 28 characters)

        Returns
        ------
            None

        Example
        -------
            Suposing calculator.memory = 10

            >>> calculator.screen()
            *------------------------------*
            |                         10.0 |
            *------------------------------*
            >>> calculator.screen("meow")
            *------------------------------*
            |             meow             |
            *------------------------------*
        """

        message = str(message).strip()
        if message:
            print(f'*{"-" * 30}*\n| {message:^28.28} |\n*{"-" * 30}*')
        else:
            print(f'*{"-" * 30}*\n| {float(self.memory):> 28} |\n*{"-" * 30}*')



import unittest
from calcoolator.handler_class import HandlerCalculator

# Default testing values
DEFAULT = 10
NEGATIVE = -2
STRING = "abc"
LIST = [1, 2, 3]
calc = HandlerCalculator()

class ExceptionsBypass(unittest.TestCase):
    """
    UnitTest to check that child class bypasses errors thrown by parent class
    """

    def test_instantiation(self):
        """
        Instantiation with wrong values defaults to 0
        """

        a = HandlerCalculator(STRING)
        b = HandlerCalculator(LIST)
        self.assertEqual(a.memory, 0)
        self.assertEqual(b.memory, 0)


    def test_add(self):
        """
        Bypass ValueError and TypeError in add() method
        and show it in STDOUT
        """
        
        calc.reset(DEFAULT)
        calc.add(3, 4)
        test = DEFAULT + 3 + 4
        self.assertEqual(calc.memory, test)

        calc.add(STRING)
        self.assertEqual(calc.memory, test)

        calc.add(LIST)
        self.assertEqual(calc.memory, test)
        
        calc.add(1, 2)
        test = test + 1 + 2
        self.assertEqual(calc.memory, test)


    def test_subtract(self):
        """
        Bypass ValueError and TypeError in subtract() method
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        calc.subtract(3, 4)
        test = DEFAULT - 3 - 4
        self.assertEqual(calc.memory, test)

        calc.subtract(STRING)
        self.assertEqual(calc.memory, test)

        calc.subtract(LIST)
        self.assertEqual(calc.memory, test)

        calc.subtract(1, 2)
        test = test - 1 - 2
        self.assertEqual(calc.memory, test)


    def test_multiply(self):
        """
        Bypass ValueError and TypeError in multiply() method
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        calc.multiply(10, -3)
        test = DEFAULT * 10 * (-3)
        self.assertEqual(calc.memory, test)

        calc.multiply(STRING)
        self.assertEqual(calc.memory, test)

        calc.multiply(LIST)
        self.assertEqual(calc.memory, test)

        calc.multiply(1, 2)
        test = test * 2
        self.assertEqual(calc.memory, test)


    def test_divide(self):
        """
        Bypass ValueError, TypeError and ZeroDivisionError in divide()
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        calc.divide(2, -1)
        test = DEFAULT / 2 / (-1)
        self.assertEqual(calc.memory, test)

        calc.divide(STRING)
        self.assertEqual(calc.memory, test)

        calc.divide(LIST)
        self.assertEqual(calc.memory, test)

        calc.divide(0)
        self.assertEqual(calc.memory, test)

        calc.divide(1, 2)
        test = test / 1 / 2
        self.assertEqual(calc.memory, test)

    def test_root(self):
        """
        Bypass all errors in nth_root() method
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        self.assertEqual(calc.memory, DEFAULT)

        calc.nth_root(STRING) # ValueError
        self.assertEqual(calc.memory, DEFAULT)

        calc.nth_root(LIST)   # TypeError
        self.assertEqual(calc.memory, DEFAULT)

        calc.nth_root(0)      # ZeroDivisionError
        self.assertEqual(calc.memory, DEFAULT)

        calc.reset(NEGATIVE)
        calc.nth_root(2)      # NegativeRootError
        self.assertEqual(calc.memory, NEGATIVE)

    def test_root(self):
        """
        Bypass all errors in power() method
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        self.assertEqual(calc.memory, DEFAULT)

        calc.power(STRING)
        self.assertEqual(calc.memory, DEFAULT)

        calc.power(LIST)
        self.assertEqual(calc.memory, DEFAULT)

        calc.reset(0)
        calc.power(NEGATIVE)
        self.assertEqual(calc.memory, 0)

    def test_reset(self):
        """
        Bypass TypeError in reset() method, set memory to 0
        and show message in STDOUT
        """

        calc.reset(DEFAULT)
        calc.reset(STRING)
        self.assertEqual(calc.memory, 0)

        calc.reset(DEFAULT)
        calc.reset(LIST)
        self.assertEqual(calc.memory, 0)

        print(calc.screen.__doc__)



if __name__ == '__main__':
    unittest.main()


from calcoolator.base_class import Calculator

DEFAULT = 5
calc = Calculator(DEFAULT)


# ========== multiply() ==========

# Regular
calc.reset(DEFAULT)
calc.multiply()
test = DEFAULT * 10 * 20 * 30
assert(calc.memory == DEFAULT)
print("Test empty args passed")

# Regular
calc.reset(DEFAULT)
calc.multiply(10, 20, 30)
test = DEFAULT * 10 * 20 * 30
assert(calc.memory == test)
print("Test regular passed")

# Parse
calc.reset(DEFAULT)
calc.multiply(10, "20", 30)
test = DEFAULT * 10 * 20 * 30
assert(calc.memory == test)
print("Test parsing positive passed")

# Parse
calc.reset(DEFAULT)
calc.multiply(10, "20", '-30')
test = DEFAULT * 10 * 20 * -30
assert(calc.memory == test)
print("Test parsing negative passed")

# Only
calc.reset(DEFAULT)
calc.multiply(10, 20, 30, only=True)
test = 10 * 20 * 30
assert(calc.memory == test)
print("Test only passed")

# Factor
calc.reset(DEFAULT)
calc.multiply(10, 20, 30, factor=True)
test = (DEFAULT*10) * (DEFAULT*20) * (DEFAULT*30)
assert(calc.memory == test)
print("Test factor passed")

# Factor
calc.reset(DEFAULT)
calc.multiply(-10, 20, 30, factor=True)
test = (DEFAULT*-10) * (DEFAULT*20) * (DEFAULT*30)
assert(calc.memory == test)
print("Test factor negative passed")

# Factor + only
calc.reset(DEFAULT)
calc.multiply(10, 20, 30, only=True, factor=True)
test = (DEFAULT*10) * (DEFAULT*20) * (DEFAULT*30)
assert(calc.memory == test)
print("Test override passed")

# Error
calc.reset(DEFAULT)
try:
    calc.multiply(10, "x", 20, 30, only=True)
except ValueError:
    print("ValueError was raised")
assert(calc.memory == DEFAULT)
assert(calc.memory != 10*20*30)
print("Test ValueError passed")

# Error
calc.reset(DEFAULT)
try:
    calc.multiply(["$", 5], "20", 30)
except TypeError:
    print("TypeError was raised")
assert(calc.memory == DEFAULT)
print("Test TypeError passed")


from base_class import Calculator
from handler_class import HandlerCalculator

base = Calculator(10)
handler = HandlerCalculator()

base.screen()
base.divide
base.screen()


# handler.screen()
# handler.reset(5)
# print(handler.memory)
# handler.screen()
# handler.add(4, 5, only=True)
# print(handler.memory)
# handler.add(5, 5, 5)
# handler.screen()
# handler.nth_root(2, -5)


import unittest
from calcoolator.base_class import Calculator

# Setting default test values
DEFAULT = 10
calc = Calculator()

class TestAdd(unittest.TestCase):
    """
    Tests for the Calculator.add() method
    """
    
    def test_single_positive(self):
        """
        Test for a single positive integer value
        """

        calc.reset(DEFAULT)
        calc.add(5)
        test = DEFAULT + 5
        self.assertEqual(calc.memory, test)

    def test_single_negative(self):
        """
        Test for a single negative integer value
        """

        calc.reset(DEFAULT)
        calc.add(-78)
        test = DEFAULT + (-78)
        self.assertEqual(calc.memory, test)

    def test_positives(self):
        """
        Test for multiple positive float values
        """

        calc.reset(DEFAULT)
        calc.add(9.5)
        calc.add(6.2)
        test = DEFAULT + 9.5 + 6.2
        self.assertEqual(calc.memory, test)

    def test_negatives(self):
        """
        Test for multiple negative float values
        """

        calc.reset(DEFAULT)
        calc.add(-10.2)
        calc.add(-4)
        test = DEFAULT + (-10.2) + (-4)
        self.assertEqual(calc.memory, test)

    def test_strings(self):
        """
        Test for checking parsing of strings
        """

        calc.reset(DEFAULT)
        calc.add(" 11  ")        # This value should be fine
        calc.add("-10")
        with self.assertRaises(ValueError):
            calc.add("   1  z")  # This value should raise ValueError
        calc.add("22")
        with self.assertRaises(ValueError):
            calc.add(" 50  x")   # Also this one
        calc.add("-.5")
        test1 = DEFAULT + 11 + (-10) + 22 + (-.5)
        test2 = DEFAULT + 11 + (-10) + 1 + 22 + 50 + (-.5)
        self.assertEqual(calc.memory, test1)      # 11-10+22-0.5 = 22.5
        self.assertNotEqual(calc.memory, test2)

    def test_big(self):
        """
        Test for checking large numbers
        """
        calc.reset(DEFAULT)
        calc.add(7908)
        calc.add(42687)
        test = DEFAULT + 7908 + 42687
        self.assertEqual(calc.memory, test)

    def test_exceptions(self):
        """
        Test for checking exceptions handling
        """

        calc.reset(DEFAULT)
        calc.add(1, 2, 3)
        with self.assertRaises(ValueError):
            calc.add("g")
        with self.assertRaises(NameError):   # Cannot be handled inside class
            calc.add(x)
        with self.assertRaises(TypeError):   # Cannot be handled inside class
            calc.add([1, 2, 3])
        test = DEFAULT + 1 + 2 + 3
        self.assertEqual(calc.memory, test)   # Memory hasn't changed



if __name__ == '__main__':
    unittest.main()


class NegativeRootError(Exception):

    def __init__(self, n, value):
        self.n = n
        self.value = value
        self.message = f"{n}√{value} I don't undertand complex numbers, sorry 💅🏻"
        super().__init__(self.message)


"""
calcoolator package info:

In this package you'll find two main modules:
    > base_class
    > handler_class

And the submodule exceptions, containing the negative_root module, a custom
    exception class created to handle roots of negative numbers as this
    calculators are not so smart and aren't designed to handle complex numbers.

Below is a summary of the basic features:

    1) base_class contains Calculator: A simple, yet effective class to compute
        basic arithmetic operations, such as addition, subtraction,
        multiplication, division, nth roots, powers and some more.
    2) handler_class contains HandlerCalculator: A child to Calculator. Same
        functionality with the difference that catches errors and continues
        operations instead of stopping execution.

For easy use, please structure your imports as follow:

    from calcoolator.base_class import Calculator
    or
    from calcoolator.handler_class import HandlerCalculator

And just create the instances in your project:

    instance = Calculator()
    or
    instance = HandlerCalculator()

For full feature description please refer to the specific help() or __doc__
    of the class or method you are interested in.

Thank you for your download! :)
"""



import unittest
from calcoolator.base_class import Calculator


DEFAULT = 10
calc = Calculator()

class TestDivide(unittest.TestCase):
    """
    Tests for the Calculator.divide() method
    """

    def test_empty(self):
        """
        Test for checking empty arguments
        """

        calc.reset(DEFAULT)
        calc.divide()
        self.assertEqual(calc.memory, DEFAULT)


    def test_single_positive(self):
        """
        Test for checking a single value
        """

        calc.reset(DEFAULT)
        calc.divide(5)
        test1 = DEFAULT / 5
        self.assertEqual(calc.memory, test1)
        calc.divide(3, reverse=True)
        test2 = 3 / test1
        self.assertEqual(calc.memory, test2)


    def test_single_negative(self):
        """
        Test for checking chained operations
        """
        
        calc.reset(DEFAULT)
        calc.divide(-5)
        test = DEFAULT / (-5)
        self.assertEqual(calc.memory, test)


    def test_multiple(self):
        """
        Test for checking chained operations
        """

        calc.reset(DEFAULT)
        calc.divide(7, 4, 8)
        calc.divide(2, 5)
        calc.divide(10, reverse=True)
        test1 = DEFAULT / 7 / 4 / 8
        test2 = test1 / 2 / 5
        test3 = 10 / test2
        self.assertEqual(calc.memory, test3)


    def test_only(self):
        """
        Test for checking "only"
        """

        calc.reset(DEFAULT)
        calc.divide(7, 4, 8, only=True)
        test1 = DEFAULT / 7 / 4 / 8
        test2 = 7 / 4 / 8
        self.assertNotEqual(calc.memory, test1)
        self.assertEqual(calc.memory, test2)


    def test_reverse(self):
        """
        Test for checking "reverse"
        """

        calc.reset(DEFAULT)
        calc.divide(7, 4, 8, reverse=True)
        test = 8 / 4 / 7 / DEFAULT
        self.assertEqual(calc.memory, test)


    def test_only_reverse(self):
        """
        Test for checking "reverse" and "only"
        """
                
        calc.reset(DEFAULT)
        calc.divide("3", -2, 5, only=True, reverse=True)
        test1 = DEFAULT / 3 / (-2) / 5              # Regular 
        test2 = 5 / (-2) / 3 / DEFAULT              # Reverse
        test3 = 3 / (-2) / 5                        # Only 
        test4 = 5 / (-2) / 3                        # Reverse & only
        self.assertNotEqual(calc.memory, test1)
        self.assertNotEqual(calc.memory, test2)
        self.assertNotEqual(calc.memory, test3)
        self.assertEqual(calc.memory, test4)


    def test_exceptions(self):
        """
        Test for checking exceptions
        """

        calc.reset(DEFAULT)
        calc.divide(1, 2, 3)
        with self.assertRaises(ValueError):
            calc.divide("meow")
        with self.assertRaises(TypeError):
            calc.divide([1, 2, 3])
        with self.assertRaises(ZeroDivisionError):
            calc.divide(6, 14, "0")
        test = DEFAULT / 1 / 2 / 3
        self.assertEqual(calc.memory, test)   # Memory hasn't changed


if __name__ == '__main__':
    unittest.main()


import unittest
from calcoolator.base_class import Calculator


# Default testing values
DEFAULT = 10
NEGATIVE = -3
calc = Calculator()

class TestPower(unittest.TestCase):
    """
    Tests for the Calculator.reset() method
    """

    def test_empty(self):
        """
        Test for resetting to 0
        """

        calc.reset()
        self.assertEqual(calc.memory, 0)

    def test_errors(self):
        """
        Test for resetting to 0
        """

        calc.reset(DEFAULT)
        with self.assertRaises(TypeError):
            calc.reset("hello")
        with self.assertRaises(TypeError):
            calc.reset("34")
        with self.assertRaises(TypeError):
            calc.reset([])
        calc.reset(NEGATIVE)
        self.assertEqual(calc.memory, NEGATIVE)


if __name__ == '__main__':
    unittest.main()


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README_pypi.md").read_text(encoding="utf-8")

setup(
    name="calcoolator",
    version="0.0.5",
    description="Simple calculator project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/berto-polo",
    author="Alberto Polo",
    author_email="bertopolo@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="calculator, newbie",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.8",
    project_urls={  # Optional
        "GitHub": "https://github.com/berto-polo",
    },
)


import unittest
from calcoolator.base_class import Calculator
from calcoolator.exceptions import NegativeRootError

# Setting default test values
DEFAULT = 10
NEGATIVE = -3
calc = Calculator()

class TestRoot(unittest.TestCase):
    """
    Tests for the Calculator.nth_root() method
    """
    
    def test_positive(self):
        """
        Test for a positive integer value
        """

        calc.reset(DEFAULT)
        calc.nth_root(2)
        test = DEFAULT ** (1/2)
        self.assertEqual(calc.memory, test)

    def test_float(self):
        """
        Test for a positive float value
        """

        calc.reset(DEFAULT)
        calc.nth_root(2.4987)
        test = DEFAULT ** (1/2)
        self.assertEqual(calc.memory, test)

    def test_negative_root(self):
        """
        Test for a negative root value
        """

        calc.reset(DEFAULT)
        calc.nth_root(NEGATIVE)
        test = DEFAULT ** (1/NEGATIVE)
        self.assertEqual(calc.memory, test)

    def test_negative_radicand(self):
        """
        Test for a negative radicand value
        """

        calc.reset(NEGATIVE)
        with self.assertRaises(NegativeRootError):
            calc.nth_root(2)
        self.assertEqual(calc.memory, NEGATIVE)

    def test_empty_positive(self):
        """
        Test for empty arguments on positive memory
        """

        calc.reset(DEFAULT)
        calc.nth_root()
        self.assertEqual(calc.memory, DEFAULT)

    def test_empty_negative(self):
        """
        Test for empty arguments on negative memory
        """

        calc.reset(NEGATIVE)
        with self.assertRaises(NegativeRootError):
            calc.nth_root()
        self.assertEqual(calc.memory, NEGATIVE)

    def test_0th_root(self):
        """
        Test for 0th root
        """

        calc.reset(DEFAULT)
        with self.assertRaises(ZeroDivisionError):
            calc.nth_root(0)
        self.assertEqual(calc.memory, DEFAULT)

    def test_exceptions(self):
        """
        Test for TypeError and ValueError
        """

        calc.reset(DEFAULT)
        with self.assertRaises(ValueError):
            calc.nth_root("a")
        with self.assertRaises(TypeError):
            calc.nth_root([1, 2, 3])
        self.assertEqual(calc.memory, DEFAULT)


if __name__ == '__main__':
    unittest.main()




DEFAULT = 5
calc = Calculator(DEFAULT)


# ========== subtract() ==========

# Regular
calc.reset(DEFAULT)
calc.subtract()
assert(calc.memory == DEFAULT)

# Regular
calc.reset(DEFAULT)
calc.subtract(1, 2, 3)
test = DEFAULT - 1 - 2 - 3
assert(calc.memory == test)

# Parse
calc.reset(DEFAULT)
calc.subtract(1, "-2", 3)
test = DEFAULT - 1 - (-2) - 3
assert(calc.memory == test)

# Only
calc.reset(DEFAULT)
calc.subtract(1, 2, 3, only=True)
test = 1 - 2 - 3
assert(calc.memory == test)

# Reverse
calc.reset(DEFAULT)
calc.subtract(1, 2, 3, reverse=True)
test = 3 - 2 - 1 - DEFAULT
assert(calc.memory == test)

# Only & reverse
calc.reset(DEFAULT)
calc.subtract(1, 2, 3, reverse=True, only=True)
test = 3 - 2 - 1
assert(calc.memory == test)

# Error
calc.reset(DEFAULT)
calc.subtract(1, "2x", 3, reverse=True)
test = 3 - 2 - 1 - DEFAULT
assert(calc.memory != test)
assert(calc.memory == DEFAULT)

# Error
calc.reset(DEFAULT)
calc.subtract(1, 2, 3, ())
assert(calc.memory != test)
assert(calc.memory == DEFAULT)


import unittest
from calcoolator.base_class import Calculator

# Setting default test values
DEFAULT = 10
POSITIVE = 2
NEGATIVE = -3
calc = Calculator()

class TestPower(unittest.TestCase):
    """
    Tests for the Calculator.power() method
    """
    
    def test_empty(self):
        """
        Test for a positive integer value
        """

        calc.reset(DEFAULT)
        calc.power()
        self.assertEqual(calc.memory, DEFAULT)
    
    def test_parsing(self):
        """
        Test for checking parsing to int
        """

        calc.reset(DEFAULT)
        calc.power(5.43)
        test = DEFAULT ** 5
        self.assertEqual(calc.memory, test)

    def test_positive_power(self):
        """
        Test for a positive integer value
        """

        calc.reset(DEFAULT)
        calc.power(POSITIVE)
        test = DEFAULT ** POSITIVE
        self.assertEqual(calc.memory, test)

    def test_negative_power(self):
        """
        Test for a positive integer value
        """

        calc.reset(DEFAULT)
        calc.power(NEGATIVE)
        test = DEFAULT ** NEGATIVE
        self.assertEqual(calc.memory, test)

    def test_override(self):
        """
        Test for override of current memory
        """

        calc.reset(DEFAULT)
        calc.power(5, 8)
        test = 8 ** 5
        self.assertEqual(calc.memory, test)

    def test_zero(self):
        """
        Test for rising 0 to a positive number
        """

        calc.reset(0)
        calc.power(POSITIVE)
        test = 0 ** POSITIVE
        self.assertEqual(calc.memory, test)

    def test_zero_negative(self):
        """
        Test for rising 0 to a negative number
        """

        calc.reset(0)
        with self.assertRaises(ZeroDivisionError):
            calc.power(NEGATIVE)
        self.assertEqual(calc.memory, 0)

    def test_zero_zero(self):
        """
        Test for raising 0 to 0
        """

        calc.reset(0)
        calc.power(0)
        self.assertTrue(0 ** 0, calc.memory)
        self.assertTrue(0 ** 0, 1)
        self.assertEqual(calc.memory, 1)

    def test_exceptions(self):
        """
        Test for checking exceptions raised
        """

        calc.reset(DEFAULT)
        with self.assertRaises(ValueError):
            calc.power("hello")
        with self.assertRaises(TypeError):
            calc.power([1, 2, 3])
        self.assertEqual(calc.memory, DEFAULT)



if __name__ == '__main__':
    unittest.main()